// Fungimol - an extensible system for designing atomic-scale objects.
// Copyright (C) 2000 Tim Freeman
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Library General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Library General Public License for more details.
// 
// You should have received a copy of the GNU Library General Public
// License along with this library in the file COPYING.txt; if not,
// write to the Free Software Foundation, Inc., 59 Temple Place -
// Suite 330, Boston, MA 02111-1307, USA
//
// The author can be reached by email at tim@fungible.com, or by
// paper mail at:
//
// Tim Freeman
// 381 N Fernwood Circle
// Sunnyvale, CA 94085-3059
//
#ifndef Integrator_h_included
#include "Integrator.h"
#endif

#ifndef SP_h_included
#include "SP.h"
#endif

#ifndef FactoryTable_h_included
#include "FactoryTable.h"
#endif

#ifndef FString_h_included
#include "FString.h"
#endif

#ifndef myassert_h_included
#include "myassert.h"
#endif

#ifndef Evaluator_h_included
#include "Evaluator.h"
#endif

#ifndef Factory_h_included
#include "Factory.h"
#endif

#ifndef TypedFactory_h_included
#include "TypedFactory.h"
#endif

#ifndef SimpleConfiguration_h_included
#include "SimpleConfiguration.h"
#endif

#ifndef iostream_h_included
#include <iostream>
#define iostream_h_included
#endif

namespace {
  // Adams Bashforth predictor, second order.
  // My reference for this is Introduction to Numerical Analysis, Second
  // Edition, by Stoer and Bulirsch, Texts in Applied Mathematics 12,  1993,
  // Springer-Verlag, pages 456-457.
  // I want a second order technique because we're going to use this to
  // integrate (among other things) the velocity as a consequence of Brenner's
  // force field.  Those forces are the derivative with respect to position of
  // the energy, and the energy function has terms that have constant functions
  // spliced together with sine functions.  The first derivative exists and
  // there is a second derivative almost everywhere, but
  // the third derivative is unbounded because the second derivative is
  // discontinuous.  So how many derivatives should we expect the velocity to
  // have?  Velocity is the integral of acceleration (that's one), and since
  // mass is constant, 
  // a = dv/dt
  // da/dt = d(dv/dt)/dt = (by chain rule) (d(dv/dt)/dx) (dx/dt)
  // dx/dt is v.  dv/dt is a.  a = F/m = dE/dx/m, so
  // da/dt = d2E/dx2/m v.  v is well behaved, and by the description above
  // d2E/dx2 is well behaved, so da/dt is well behaved.
  // d2a/dt2 = d(da/dt)/dt = d(d2E/dx2/m)/dt v + d2E/dx2/m dv/dt
  //         = d(d2E/dx2/m)/dx dx/dt v + d2E/dx2/m a
  //         = d3E/dx3/m v**2 + d2E/dx2/m a
  // As discussed earlier, d3E/dx3 is unbounded, so d2a/dt2 is unbounded.
  // Thus we only have three bounded derivatives of position: v, a, and da/dt.
  // But worse, we only have two bounded derivatives of velocity.  
  // 
  // The beauty of the Nordsieck technique is that it does position directly in
  // terms of acceleration, so you have three derivatives and can increase the
  // order by one.  Also claims to be a predictor-corrector method that
  // requires only one evaluation of the function per timestep.  But
  // DesignAtoms want to do something other than position in terms of
  // acceleration, so we can't use Nordsieck throughout.  I don't know yet
  // whether the additional higher order matters; if it does, then I suppose I
  // should apply different techniques to different coordinates.  But I can't
  // do that because all of the techniques other than Nordsieck want to
  // evaluate the derivatives at past estimated positions.  Nordsieck, in
  // contrast, never evaluates the derivatives at a final estimation of a
  // position; it always takes an estimated position, predicts, evaluates the
  // derivatives, and then corrects to get the next estimated position.
  class AB2 : public Integrator {
    int m_startedSize;
    FFloat m_time;
    FFloat m_step;
    // m_k0 is history, the derivative at the previous timestep, or null if we
    // haven't accumulated history yet.
    FFloat *m_k0;
    // m_deriv is logically local to observe, but I keep it global to avoid
    // repeated memory allocation in each timestep.
    FFloat *m_deriv;
    SP<Evaluator> m_f;
    // m_rk is a RungeKutta evaluator so we can get our first few timesteps.
    SP<Integrator> m_rk;
    // m_started is a boolean flag saying whether we're started.  An integrity
    // check.
    bool m_started;
    bool isStarted ();
    void deleteAll ();
  public:
    AB2 ();
    void start (FFloat initialTime, FFloat initialStep);
    void stop (FFloat &time, FFloat &step);
    void setEvaluator (Evaluator *f);
    Evaluator *getEvaluator ();
    FFloat observe (const FFloat *deriv);
    ~AB2 ();
  };

  AB2::AB2 () {
    SP<Configurable> rk =
      FactoryTable::load ("Integrator", "RungeKutta")->makeIt ();
    assert (dynamic_cast <Integrator *> (&*rk));
    m_rk = static_cast <Integrator *> (&*rk);
    m_k0 = 0;
    m_deriv = 0;
  }

  static inline void deleteIfNonzero (FFloat *&f) {
    if (f) {
      delete [] f;
      f = 0;
    }
  }

  void AB2::deleteAll () {
    deleteIfNonzero (m_k0);
    deleteIfNonzero (m_deriv);
  }

  AB2::~AB2 () {
    deleteAll ();
  }

  void AB2::start (FFloat initialTime, FFloat initialStep) {
    assert (0 == m_deriv && 0 == m_k0);
    m_time = initialTime;
    m_step = initialStep;
    assert (m_f);
    m_startedSize = m_f->stateVec().size();
    m_started = true;
    m_deriv = NEW (FFloat [m_startedSize]);
    // Leave m_k0 zero to reflect the fact that we don't have any history yet.
  }

  void AB2::setEvaluator (Evaluator *f) {
    m_f = f;
    m_rk->setEvaluator (f);
  }

  Evaluator *AB2::getEvaluator () {
    return m_f;
  }

  bool AB2::isStarted () {
    return m_started;
  }

  void AB2::stop (FFloat &time, FFloat &step) {
    time = m_time;
    step = m_step;
    assert (isStarted ());
    deleteAll ();
    m_started = false;
  }

  FFloat AB2::observe (const FFloat *deriv) {
    assert (isStarted ());
    // Same issues with retrying because of things deleting themselves in the
    // middle as with RungeKutta. 
    for (;;) {
      if (m_f->stateVec().size() != m_startedSize) {
        FFloat time, step;
        stop (time, step);
        start (time, step);
        deriv = 0;
      }
      FFloat *state = &(m_f->stateVec()[0]);
      bool retry;
      if (deriv) {
        retry = false;
        for (int i = 0; i < m_startedSize; i++) {
          m_deriv [i] = deriv [i];
        }
        deriv = 0;
      } else {        
        retry = m_f->eval (m_time, state, m_deriv);
      }
      if (retry) {
        // Go around again
      } else if (0 == m_k0) {
        // Not enough history; do a RungeKutta step.
        m_rk->start (m_time, m_step);
        // FIXME m_rk will do a call to eval with the same arguments that we're
        // going to call it now, in the normal case where no retry happens due
        // to atoms falling off the edge of space.  Wasted effort.  However,
        // it's only wasted at startup time.
        const FFloat time = m_rk->observe (m_deriv);
        (void) time;
        {
          FFloat junk1, junk2;
          m_rk->stop (junk1, junk2);
        }
        assert (m_time + m_step == time);
        if (m_f->stateVec().size() != m_startedSize) {
          // m_rk->observe must have deleted an atom.
          retry = true;
        } else {
          m_k0 = m_deriv;
          m_deriv = NEW (FFloat [m_startedSize]);
        }
      } else {
        // Enough history; do a predictor step.
        for (int i = 0; i < m_startedSize; i++) {
          state [i] = state[i] + m_step*((3.0/2.0)*m_deriv[i]+
                                         (-1.0/2.0)*m_k0[i]);
        }
        // Cycle the history.
        {
          FFloat * cycle = m_k0;
          m_k0 = m_deriv;
          m_deriv = cycle;
          // Now the values in m_deriv will be overwritten before they are next
          // used; we're just keeping it around so we don't have to deallocate
          // it and then reallocate it next time.
        }
      }
      if (!retry) {
        break;
      }
    }
    m_time += m_step;
    return m_time;
  }

  class AB2Factory :
    public TypedFactory <SimpleConfiguration, AB2>
  {
  public:
    AB2Factory () :
      TypedFactory <SimpleConfiguration, AB2> ("AB2")
    {}
    SP<SimpleConfiguration> typedDefaultConfiguration () const {
      return new SimpleConfiguration ();
    }
    SP<AB2> makeIt (SimpleConfiguration *c) const {
      (void) c;
      return NEW (AB2 ());
    }
  };
  
  bool useless =
  (FactoryTable::store ("Integrator", NEW (AB2Factory ())),
   true);
}
