// Fungimol - an extensible system for designing atomic-scale objects.
// Copyright (C) 2000 Tim Freeman
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Library General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Library General Public License for more details.
// 
// You should have received a copy of the GNU Library General Public
// License along with this library in the file COPYING.txt; if not,
// write to the Free Software Foundation, Inc., 59 Temple Place -
// Suite 330, Boston, MA 02111-1307, USA
//
// The author can be reached by email at tim@fungible.com, or by
// paper mail at:
//
// Tim Freeman
// 381 N Fernwood Circle
// Sunnyvale, CA 94085-3059
//

#ifndef FFloat_h_included
#include "FFloat.h"
#endif

#ifndef Evaluator_h_included
#include "Evaluator.h"
#endif

#ifndef FactoryTable_h_included
#include "FactoryTable.h"
#endif

#ifndef myassert_h_included
#include "myassert.h"
#endif
  
#ifndef Configuration_h_included
#include "Configuration.h"
#endif

#ifndef SP_h_included
#include "SP.h"
#endif

#ifndef FString_h_included
#include "FString.h"
#endif

#ifndef MemoryUtil_h_included
#include "MemoryUtil.h"
#endif

#ifndef TypedFactory_h_included
#include "TypedFactory.h"
#endif

#ifndef Integrator_h_included
#include "Integrator.h"
#endif

#ifndef SimpleConfiguration_h_included
#include "SimpleConfiguration.h"
#endif

#ifndef iostream_h_included
#include <iostream>
#define iostream_h_included
#endif

namespace {
  // From Numerical Analysis, Second Edition, by Burden, Faires, and
  // Reynolds, 1978, page 205ff.
  class RungeKutta : public Integrator {
    int m_startedSize;
    FFloat m_time;
    FFloat m_step;
    // Next ones are allocated and deleted by this class.
    // Beware -- my m_k1 * m_step is equal to their K1.
    FFloat *m_k1;
    FFloat *m_k2;
    FFloat *m_k3;
    FFloat *m_k4;
    FFloat *m_tmp;
    SP<Evaluator> m_f;
    bool isStarted ();
  public:
    RungeKutta ();
    void start (FFloat initialTime,
                FFloat initialStep);
    void stop (FFloat &time, FFloat &step);
    void setEvaluator (Evaluator *f);
    Evaluator *getEvaluator ();
    FFloat observe (const FFloat *deriv);
    ~RungeKutta ();
  };
  
  RungeKutta::RungeKutta () {
    m_k1 = 0;
  }
  
  RungeKutta::~RungeKutta () {
    if (m_k1) {
      FFloat j1, j2;
      stop (j1, j2);
    }
  }
  
  void RungeKutta::start (FFloat initialTime,
                          FFloat initialStep) {
    assert (0 == m_k1);
    m_time = initialTime;
    m_step = initialStep;
    assert (m_f);
    m_startedSize = m_f->stateVec().size();
    m_k1 = NEW (FFloat [m_startedSize]);
    m_k2 = NEW (FFloat [m_startedSize]);
    m_k3 = NEW (FFloat [m_startedSize]);
    m_k4 = NEW (FFloat [m_startedSize]);
    m_tmp = NEW (FFloat [m_startedSize]);
  }
  
  void RungeKutta::setEvaluator (Evaluator *f) {
    m_f = f;
  }
  
  Evaluator *RungeKutta::getEvaluator () {
    return m_f;
  }
  
  bool RungeKutta::isStarted () {
    if (m_k1) {
      assert (m_k2 && m_k3 && m_k4 && m_tmp);
      return true;
    } else {
      return false;
    }
  }
  
  void RungeKutta::stop (FFloat &time, FFloat &step) {
    time = m_time;
    step = m_step;
    assert (isStarted());
    delete [] m_k1;
    delete [] m_k2;
    delete [] m_k3;
    delete [] m_k4;
    delete [] m_tmp;
    m_k1 = 0;
  }
  
  FFloat RungeKutta::observe (const FFloat *deriv) {
    assert (isStarted ());
    // If an object is deleted in the middle of a timestep, then its
    // derivative will be zero.  This discontinuity doesn't faze
    // RungeKutta because it isn't trying to determine whether it is
    // being accurate or not.  However, it probably could cause an
    // adaptive algorithm to have fits, so let's try catching the
    // exception and doing something constructive with it now.
    for (;;) {
      if (m_f->stateVec().size() != m_startedSize) {
        // The size of the state changed, so we have to reallocate the arrays.
        FFloat time, step;
        stop (time, step);
        start (time, step);
        deriv = 0;
      }
      FFloat * state = &(m_f->stateVec()[0]);
      bool retry;
      if (deriv) {
        retry = false;
        for (int i = 0; i < m_startedSize; i++) {
          m_k1 [i] = deriv [i];
        }
        deriv = 0;
      } else {
        retry = m_f->eval (m_time, state, m_k1);
      }
      if (!retry) {
        // FIXME SIMD instructions would be great for each of these loops.
        // Prefetches might help too.
        for (int i = 0; i < m_startedSize; i++) {
          m_tmp [i] = state [i] + m_step * m_k1 [i] / 2;
        }
        retry = m_f->eval (m_time + m_step / 2, m_tmp, m_k2);
      }
      if (!retry) {
        for (int i = 0; i < m_startedSize; i++) {
          m_tmp [i] = state [i] + m_step * m_k2 [i] / 2;
        }
        retry = m_f->eval (m_time + m_step / 2, m_tmp, m_k3);
      }
      if (!retry) {
        for (int i = 0; i < m_startedSize; i++) {
          m_tmp [i] = state [i] + m_step * m_k3 [i];
        }
        retry = m_f->eval (m_time + m_step, m_tmp, m_k4);
      }
      if (!retry) {
        for (int i = 0; i < m_startedSize; i++) {
          state [i] = state [i] +
            m_step * (m_k1 [i] + 2 * m_k2 [i] + 2 * m_k3 [i] + m_k4 [i]) / 6;
        }
      }
      if (!retry) {
        break;
      }
    }
    m_time += m_step;
    return m_time;
  }
  
  class RungeKuttaFactory :
    public TypedFactory <SimpleConfiguration, RungeKutta>
  {
  public:
    RungeKuttaFactory () :
      TypedFactory <SimpleConfiguration, RungeKutta> ("RungeKutta")
    {}
    SP<SimpleConfiguration> typedDefaultConfiguration () const {
      return new SimpleConfiguration ();
    }
    SP<RungeKutta> makeIt (SimpleConfiguration *c) const {
      (void) c;
      return NEW (RungeKutta ());
    }
  };
  
  bool useless =
  (FactoryTable::store ("Integrator", NEW (RungeKuttaFactory ())),
   true);
}

