#!/usr/local/sbin/perl -w
use strict;
my $default_threshold = 0.000002;
sub usage () {
    die << "EOM";
Usage: $0 file1.d file2.d <threshold>
Scan through these files (which are the same format as the xmol.d
files from brennermd) and report any differences.  A difference is two
corresponding coordinates that differ by more than the threshold. 

The default value for threshold is $default_threshold.
EOM
}

my $threshold;

sub docmp ($$) {
    my ($n1, $n2) = @_;
    if (abs ($n1 - $n2) > $threshold) {
	die "Significantly different numbers $n1 and $n2 found";
    }
}

sub main () {
    usage () unless 3 == @ARGV || 2 == @ARGV;
    my ($file1, $file2) = @ARGV;
    if (3 == @ARGV) {
	$threshold = $ARGV[2];
    } else {
	$threshold = $default_threshold;
    }
    open (FILE1, "<$file1") || die "Can't read $file1: $!";
    open (FILE2, "<$file2") || die "Can't read $file2: $!";
    my $sum_diff = 0.0;
    my $max_dist = 0.0;
    my $atomcount;
    my ($xmin, $xmax, $ymin, $ymax, $zmin, $zmax);
    for (;;) {
	my $line1 = <FILE1>;
	my $line2 = <FILE2>;
	last if !defined ($line1) && !defined ($line2);
	die "The files have different numbers of lines"
	    if !defined ($line1) || !defined ($line2);

	# It would have been cleaner to use object oriented perl to
	# eliminate the redundant code here and the code here that is
	# redundant with normalize_d.pl.  But if this is all the
	# parsing of xmol files I have to do, cut-and-paste is faster
	# to write.  I promise myself I'll rewrite this to be
	# object-oriented if we do this one more time. 
	my $countline1 = <FILE1>;
	$countline1 =~ m!^\s*(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s*$!
	    || die "Bad line $countline1 from $file1";
	my $atomcount1 = $1;
	$atomcount = $atomcount1;
	my $countline2 = <FILE2>;
	$countline2 =~ m!^\s*(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s*$!
	    || die "Bad line $countline1 from $file2";
	my $atomcount2 = $1;
	die "Different numbers of atoms"
	    if $atomcount1 != $atomcount2;

	my $timeline1 = <FILE1>;
	$timeline1 =~ m!^\s*([^\s]+)\s+([^\s]+)\s*$!
	    || die "Bad line $timeline1 from $file1";
	my ($t11, $t21) = ($1, $2);
	my $timeline2 = <FILE2>;
	$timeline2 =~ m!^\s*([^\s]+)\s+([^\s]+)\s*$!
	    || die "Bad line $timeline2 from $file2";
	my ($t12, $t22) = ($1, $2);
	docmp ($t11, $t12); docmp ($t21, $t22);
	
	my $boxline1 = <FILE1>;
	$boxline1 =~ m!^\s*([^\s]+)\s*([^\s]+)\s+([^\s]+)\s*$!
	    || die "Bad line $boxline1 from $file1";
	my ($b11, $b21, $b31) = ($1, $2, $3);
	my $boxline2 = <FILE2>;
	$boxline2 =~ m!^\s*([^\s]+)\s*([^\s]+)\s+([^\s]+)\s*$!
	    || die "Bad line $boxline2 from $file2";
	my ($b12, $b22, $b32) = ($1, $2, $3);
	# Comparing the cubes isn't interesting, since we can compile
	# with -DINFINITE_CUBE.
	#docmp ($b11, $b12); docmp ($b21, $b22); docmp ($b31, $b32);

	my $i;
	for ($i = 1; $i <= $atomcount1; $i++) {
	    my $line1 = <FILE1>;
	    $line1 =~ m!^\s*(\d+)\s+(\d+)\s+([^\s]+)\s+([^\s]+)\s+([^\s]+)\s*$!
		|| die "Bad line $line1 from $file1";
	    my ($lineno1, $atom1, $x1, $y1, $z1) = ($1, $2, $3, $4, $5);
	    if (defined $xmin) {
		if ($xmin > $x1) {$xmin = $x1;}
		if ($xmax < $x1) {$xmax = $x1;}
		if ($ymin > $y1) {$ymin = $y1;}
		if ($ymax < $y1) {$ymax = $y1;}
		if ($zmin > $z1) {$zmin = $z1;}
		if ($zmax < $z1) {$zmax = $z1;}
	    } else {
		$xmin = $xmax = $x1;
		$ymin = $ymax = $y1;
		$zmin = $zmax = $z1;
	    }
	    die "Bad line number $lineno1 (expected $i) from $file1 in $line1"
		unless $lineno1 == $i;

	    my $line2 = <FILE2>;
	    $line2 =~ m!^\s*(\d+)\s+(\d+)\s+([^\s]+)\s+([^\s]+)\s+([^\s]+)\s*$!
		|| die "Bad line $line2 from $file2";
	    my ($lineno2, $atom2, $x2, $y2, $z2) = ($1, $2, $3, $4, $5);
	    die "Bad line number $lineno2 (expected $i) from $file2 in $line2"
		unless $lineno2 == $i;

	    docmp ($atom1, $atom2);
	    docmp ($x1, $x2); docmp ($y1, $y2); docmp ($z1, $z2);
	    my $dist = ($x2 - $x1)*($x2 - $x1) + ($y2 - $y1)*($y2 - $y1)
		+ ($z2 - $z1)*($z2 - $z1);
	    $sum_diff += $dist;
	    $max_dist = $dist if($dist > $max_dist);
	}
    }
    close (FILE1) || die "Can't close $file1: $!";
    close (FILE2) || die "Can't close $file2: $!";
    print << "EOM";
Coordinate range in $file1 is:
   X: $xmin to $xmax
   Y: $ymin to $ymax
   Z: $zmin to $zmax
EOM
    printf "Files are close enough. %.3e average difference %.3e max difference\n",
    $sum_diff/$atomcount, $max_dist;
}

main ();
